library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
entity LVDS is
    Port ( clk : in std_logic;
           clk133 : in std_logic;
           hq : in std_logic_vector(19 downto 0);
           cc_rst : in std_logic;
           gc_block : in std_logic;
           gc_tstrip : in std_logic;
           gc_idmatch : in std_logic;
           gc_efifo : in std_logic;
           gc_automon : in std_logic;
           inh : in std_logic;
           cs_full : in std_logic;
           cs_over : in std_logic;
           ch_number : in std_logic_vector(1 downto 0);
           fqw : out std_logic_vector(32 downto 0);
           lfifo_read : in std_logic;
		   trailer : out std_logic;
		   dav_lfifo : out std_logic;
		   high_lfifo : out std_logic;
-- monitoring counting signals
		   update : out std_logic;
		   WCOUNT : out std_logic;
		   BADID : out std_logic;
		   UERR : out std_logic;
		   CERR : out std_logic
			  );
end LVDS;
architecture my_arch of LVDS is
COMPONENT lpm_ram_dp
   GENERIC ( LPM_WIDTH: POSITIVE :=17;
             LPM_WIDTHAD: POSITIVE := 8;
             LPM_NUMWORDS: NATURAL := 256;
             LPM_TYPE: STRING := "LPM_RAM_DP";
             LPM_INDATA: STRING := "REGISTERED";
             LPM_OUTDATA: STRING := "REGISTERED";
             LPM_RDADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_WRADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_FILE: STRING := "UNUSED";
             LPM_HINT: STRING := "UNUSED"

           );
   PORT    ( rdaddress, wraddress: IN STD_LOGIC_VECTOR(LPM_WIDTHAD-1 DOWNTO 0);
             rdclock, wrclock: IN STD_LOGIC := '0';
             rden, rdclken, wrclken: IN STD_LOGIC := '1';
             wren: IN STD_LOGIC; 
             data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
             q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0));
END COMPONENT;
COMPONENT lpm_counter
   GENERIC (LPM_WIDTH: POSITIVE := 7;
      LPM_MODULUS: NATURAL := 0;
      LPM_DIRECTION: STRING := "UNUSED";
      LPM_AVALUE: STRING := "UNUSED";
      LPM_SVALUE: STRING := "UNUSED";
      LPM_PVALUE: STRING := "UNUSED";
      LPM_TYPE: STRING := "LPM_COUNTER";
      LPM_HINT: STRING := "UNUSED");
   PORT (data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0) := (OTHERS => '0');
      clock: IN STD_LOGIC;
      cin: IN STD_LOGIC := '0';
      clk_en: IN STD_LOGIC := '1';
      cnt_en: IN STD_LOGIC := '1';
      updown: IN STD_LOGIC := '1';
      sload: IN STD_LOGIC := '0';
      sset: IN STD_LOGIC := '0';
      sclr: IN STD_LOGIC := '0';
      aload: IN STD_LOGIC := '0';
      aset: IN STD_LOGIC := '0';
      aclr: IN STD_LOGIC := '0';
      cout: OUT STD_LOGIC := '0';
      q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0));
END COMPONENT;
type state is (idle,head,dat1,dat0,ovr0,even,trlr);
signal RCV : state;
signal e_odd : std_logic;
signal e_struct : std_logic;
signal e_trunc : std_logic;
signal e_over : std_logic;
signal bad_id : std_logic;
signal IDREG : std_logic_vector(7 downto 0);
signal erh : std_logic_vector(1 downto 0);
signal over_sync : std_logic_vector(1 downto 0);
signal full_sync : std_logic_vector(1 downto 0);
signal over : std_logic;
signal full : std_logic;
-- LVDS fifo signals
signal preg : std_logic_vector(15 downto 0);
signal block_wc : std_logic_vector(15 downto 0);
signal we_lfifo : std_logic;
signal lfifo_wa : std_logic_vector(7 downto 0);
signal lfifo_ra : std_logic_vector(7 downto 0);
signal lfifo_wc : std_logic_vector(6 downto 0);
signal lfifo_di : std_logic_vector(16 downto 0);
signal lfifo_do : std_logic_vector(16 downto 0);
signal fqw_l : std_logic_vector(15 downto 0);
signal wa_sync : std_logic_vector(3 downto 0);
signal uerr_sync : std_logic_vector(2 downto 0);
signal cerr_sync : std_logic_vector(2 downto 0);
signal erh_toggle : std_logic_vector(1 downto 0);
signal wr_odd : std_logic;
signal up_count : std_logic;
signal ec_lfifo_wc : std_logic;
signal lfifo_read_dl : std_logic_vector(1 downto 0);
signal header : std_logic;
signal trailer_i : std_logic;
signal updc : std_logic;
begin
trailer <= trailer_i;
-- signal sync cross clock domains
process(clk,cc_rst)
begin
  if(cc_rst = '1')then
    over_sync <= "00";
    full_sync <= "00";
    over <= '0';
    full <= '0';
  elsif(clk'event and clk = '1')then
    over_sync <= over_sync(0) & cs_over;
    full_sync <= full_sync(0) & cs_full;
	if(over_sync(1) = '1')then
	  over <= not gc_efifo;
	elsif(RCV = trlr)then
	  over <= '0';
	end if;
	if(full_sync(1) = '1')then
	  full <= not gc_efifo;
	elsif(RCV = trlr)then
	  full <= '0';
	end if;
  end if;
end process;
-- state machine
process(clk,cc_rst)
begin
  if(cc_rst = '1')then
    RCV <= idle;
  elsif(clk'event and clk = '1')then
    if(RCV = idle and hq(17) = '1')then
	  RCV <= head;
	end if;
    if(RCV = head)then
	  RCV <= dat1;
	end if;
    if(RCV = dat1)then
	  if(hq(16) = '1')then
	    RCV <= even;
	  elsif(hq(17) = '1')then
-- it is better to let the event finish as overflow
--		if(full = '1')then
--		  RCV <= even;
--		elsif(over = '1')then
		if(over = '1')then
		  RCV <= ovr0;
		else
		  RCV <= dat0;
		end if;
	  end if;
	end if;
    if(RCV = dat0)then
	  if(hq(16) = '1')then
	    RCV <= trlr;
	  else
		RCV <= dat1;
	  end if;
	end if;
    if(RCV = ovr0)then
	  if(hq(16) = '1')then
	    RCV <= trlr;
	  else
		RCV <= dat1;
	  end if;
	end if;
    if(RCV = even)then
	  RCV <= trlr;
	end if;
    if(RCV = trlr)then
	  if(hq(17 downto 16) = "11" and full = '0')then
	    RCV <= head;
	  else
		RCV <= idle;
	  end if;
	end if;
  end if;
end process;
-- err status
process(clk,cc_rst)
begin
  if(cc_rst = '1')then
    e_struct <= '0';
    e_trunc <= '0';
    e_odd <= '0';
    e_over <= '0';
    erh <= "00";
  elsif(clk'event and clk = '1')then
    if(RCV = trlr)then
      e_struct <= '0';
      e_trunc <= '0';
      e_odd <= '0';
      e_over <= '0';
      erh <= "00";
	else
	  if(RCV = idle and hq(17 downto 16) = "10" and full = '0')then
		e_struct <= '1';
	  end if;
	  if(RCV = head and hq(17 downto 16) /= "10")then
		e_struct <= '1';
	  end if;
	  if(RCV = dat1 and hq(16) = '1')then
		e_struct <= '1';
	  end if;
	  if(RCV = dat0 and hq(16) = hq(17))then
		e_struct <= '1';
	  end if;
	  if(RCV = ovr0 and hq(16) = hq(17))then
		e_struct <= '1';
	  end if;
	  if(RCV = ovr0 or (RCV = dat0 and hq(17 downto 16) = "10" and over = '1'))then
		e_trunc <= '1';
	  end if;
	  if(RCV = even)then
		e_odd <= '1';
	  end if;
	  if(RCV = dat1 and hq(17 downto 16) = "10" and full = '1')then
		e_over <= '1';
	  end if;
	  erh <= erh or hq(19 downto 18);
	end if;
  end if;
end process;
process(clk,cc_rst)
begin
  if(cc_rst = '1')then
    erh_toggle <= "00";
  elsif(clk'event and clk = '1')then
	erh_toggle <= erh_toggle xor hq(19 downto 18);
  end if;
end process;
process(clk133,cc_rst)
begin
  if(cc_rst = '1')then
    uerr_sync <= "000";
    cerr_sync <= "000";
    UERR <= '0';
    CERR <= '0';
  elsif(clk133'event and clk133 = '1')then
    uerr_sync <= uerr_sync(1 downto 0) & erh_toggle(1);
    cerr_sync <= cerr_sync(1 downto 0) & erh_toggle(0);
    UERR <= uerr_sync(2) xor uerr_sync(1);
    CERR <= cerr_sync(2) xor cerr_sync(1);
  end if;
end process;
-- LVDS fifo
process(clk,cc_rst)
begin
  if(cc_rst = '1')then
    we_lfifo <= '0';
    wr_odd <= '0';
  elsif(clk'event and clk = '1')then
    if(RCV = head or RCV = even or (RCV = ovr0 and hq(16) = '1') or
	  (RCV = dat0 and (over = '0' or hq(16) = '1')))then
	  we_lfifo <= '1';
      wr_odd <= '0';
	else
	  we_lfifo <= we_lfifo and not wr_odd;
      wr_odd <= wr_odd xor we_lfifo;
	end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(RCV = head)then
	   IDREG <= preg(7 downto 0);
	 end if;
    if(RCV = head)then
	   block_wc <= "0000000000000001";
	 elsif(we_lfifo = '1' and wr_odd = '0')then
	   block_wc <= block_wc + gc_block;
	 end if;
    if(RCV = head)then
	   lfifo_di(15 downto 0) <= preg(15 downto 10) & ch_number & preg(7 downto 0);
	 elsif(RCV = trlr and gc_tstrip = '0' )then
	   lfifo_di(15 downto 0) <= preg(15 downto 8) & e_odd & e_struct & e_over & '0' & bad_id & e_trunc & erh;
	 elsif((RCV = even or (HQ(16) = '1' and (RCV = ovr0 or RCV = dat0))) and gc_block = '1')then
	   lfifo_di(15 downto 0) <= block_wc;
     else
	   lfifo_di(15 downto 0) <= preg;
	 end if;
    if(RCV = trlr)then
	   lfifo_di(16) <= '1';
     else
	   lfifo_di(16) <= '0';
	 end if;
     preg <= hq(15 downto 0);
     if(hq(15 downto 8) /= IDREG and gc_idmatch = '1')then
       bad_id <= '1';
     else
       bad_id <= '0';
     end if;
  end if;
end process;
process(clk,cc_rst)
begin
  if(cc_rst = '1')then
    lfifo_wa <= (others => '0');
  elsif(clk'event and clk = '1')then
    if(we_lfifo = '1')then
	  lfifo_wa <= lfifo_wa + 1;
	end if;
  end if;
end process;
process(clk133,cc_rst)
begin
  if(cc_rst = '1')then
    wa_sync <= "0000";
	up_count <= '0';
    lfifo_ra(7 downto 1) <= "0000000";
  elsif(clk133'event and clk133 = '1')then
    wa_sync <= wa_sync(2 downto 0) & lfifo_wa(1);
	up_count <= wa_sync(3) xor wa_sync(2);
    if(lfifo_read = '1')then
	  lfifo_ra(7 downto 1) <= lfifo_ra(7 downto 1) + 1;
	end if;
  end if;
end process;
lfifo_ra(0) <= lfifo_read;
i_lfifo_wc: lpm_counter
   port map(
           clock => clk133,
           clk_en => ec_lfifo_wc,
		   updown => up_count,
           aclr => cc_rst,
           q => lfifo_wc);
ec_lfifo_wc <= up_count xor lfifo_read;
high_lfifo <= lfifo_wc(6);
i_lfifo: lpm_ram_dp
   port map(
           rdaddress => lfifo_ra,
           wraddress => lfifo_wa,
           wrclock => clk,
           rdclock => clk133,
           wren => we_lfifo,
           data => lfifo_di,
           q => lfifo_do);
WCOUNT <= lfifo_read_dl(1);
process(clk133,cc_rst)
begin
  if(cc_rst = '1')then
	lfifo_read_dl <= "00";
	dav_lfifo <= '0';
	header <= '1';
    updc <= '0';
    update <= '0';
    BADID <= '0';
	trailer_i <= '0';
  elsif(clk133'event and clk133 = '1')then
	lfifo_read_dl <= lfifo_read_dl(0) & lfifo_read;
    if(lfifo_wc(6 downto 1) = "000000" and (lfifo_wc(0) = '0' or lfifo_read = '1'))then
	  dav_lfifo <= '0';
	else
	  dav_lfifo <= (not cs_full or gc_efifo) and (lfifo_wc(6) or not inh);
	end if;
    header <= trailer_i or (header and not lfifo_read_dl(0));
-- if bit(15) of header = '1', update channel registers
    if(header = '1' and lfifo_read_dl(0) = '1')then
      updc <= lfifo_do(15);
    end if;
    update <= not gc_automon and updc and trailer_i;
    BADID <= fqw_l(3) and trailer_i;
    trailer_i <= lfifo_do(16) and lfifo_read_dl(1);
  end if;  
end process;
process(clk133)
begin
  if(clk133'event and clk133 = '1')then
    fqw_l <= lfifo_do(15 downto 0);
  end if;
end process;
fqw <= lfifo_do & fqw_l;
end my_arch;
